
class APIPerformanceTime:
    def __init__(self):
        self.project_name = None
        self.mined_entity_directory = None
        self.xml_conv_overhead_time = 0.0
        self.mining_time = 0.0
        self.tot_attr_processing_time = 0.0
        self.tot_src_files_processed = 0
        self.tot_packages_processed = 0
        self.tot_classes_processed = 0
        self.tot_enums_processed = 0
        self.tot_methods_processed = 0
        self.tot_interfaces_processed = 0
        self.tot_static_blocks_processed = 0
        self.tot_comments_processed = 0
        self.tot_kloc_processed = 0.0
        self.avg_source_file_loc = 0.0



